import os
from pathlib import Path
from unittest.mock import patch

from obspy import UTCDateTime
from obspy.io.reftek.packet import PACKET_FINAL_DTYPE

from sohstationviewer.model.mseed_data.record_reader_helper import Unpacker
from sohstationviewer.model.reftek_data.reftek_reader import soh_packet
from sohstationviewer.model.reftek_data.reftek_reader.header import (
    NotRT130FileError, PacketHeader,
)
from sohstationviewer.model.reftek_data.reftek_reader.packet_readers import (
    read_eh_et_packet, read_dt_packet, read_soh_packet,
)
from sohstationviewer.model.reftek_data.reftek_reader.packets import (
    SOHPacket, EHETPacket, DTPacket,
)
from sohstationviewer.model.reftek_data.reftek_reader.reftek_reader_helper \
    import (
        read_rt130_file, convert_waveform_packet_to_obspy_format,
        convert_packet_header_to_dict, convert_soh_packet_to_obspy_format,
    )
from tests.base_test_case import BaseTestCase

unpacker = Unpacker('>')


class TestReadRT130File(BaseTestCase):
    def setUp(self) -> None:
        self.TEST_DATA_DIR = Path(os.getcwd()).joinpath('tests/test_data')
        self.rt130_dir = self.TEST_DATA_DIR.joinpath(
            'RT130-sample/2017149.92EB/2017150/92EB'
        )

        eh_et_patcher = patch(
            'sohstationviewer.model.reftek_data.reftek_reader.'
            'reftek_reader_helper.'
            'read_eh_et_packet',
            wraps=read_eh_et_packet
        )
        self.mock_read_eh_et = eh_et_patcher.start()
        self.addCleanup(eh_et_patcher.stop)

        dt_patcher = patch(
            'sohstationviewer.model.reftek_data.reftek_reader.'
            'reftek_reader_helper.'
            'read_dt_packet',
            wraps=read_dt_packet
        )
        self.mock_read_dt = dt_patcher.start()
        self.addCleanup(dt_patcher.stop)

        soh_patcher = patch(
            'sohstationviewer.model.reftek_data.reftek_reader.'
            'reftek_reader_helper.'
            'read_soh_packet',
            wraps=read_soh_packet
        )
        self.mock_read_soh = soh_patcher.start()
        self.addCleanup(soh_patcher.stop)

    def test_rt130_soh_file(self):
        file = self.rt130_dir.joinpath('0/000000000_00000000')
        packets = read_rt130_file(file, unpacker)
        self.assertTrue(
            all(isinstance(packet, SOHPacket) for packet in packets)
        )
        self.assertFalse(self.mock_read_dt.called)
        self.assertFalse(self.mock_read_eh_et.called)
        self.assertTrue(self.mock_read_soh.called)

    def test_rt130_raw_data_file(self):
        file = self.rt130_dir.joinpath('1/000000015_0036EE80')
        packets = read_rt130_file(file, unpacker)
        self.assertTrue(all(
            isinstance(packet, EHETPacket) or isinstance(packet, DTPacket)
            for packet in packets)
        )
        self.assertTrue(self.mock_read_dt.called)
        self.assertTrue(self.mock_read_eh_et.called)
        self.assertFalse(self.mock_read_soh.called)

    def test_non_rt130_file(self):
        with self.subTest('test_file_exist'):
            file = self.TEST_DATA_DIR.joinpath(
                'Q330-sample/day_vols_AX08/AX08.XA..HHE.2021.186'
            ).as_posix()
            with self.assertRaises(NotRT130FileError):
                read_rt130_file(file, unpacker)

        with self.subTest('test_file_does_not_exist'):
            file = ''
            with self.assertRaises(FileNotFoundError):
                read_rt130_file(file, unpacker)


class TestConvertPacketHeaderToDict(BaseTestCase):
    def setUp(self):
        self.header = PacketHeader('EH', 1000, '98E1',
                                   UTCDateTime(year=1951, month=1, day=1),
                                   1024, 1)

    def test_all_fields_are_in_the_converted_dict(self):
        expected_fields = ['packet_type', 'experiment_number', 'year',
                           'unit_id', 'time', 'byte_count', 'packet_sequence']
        actual_field = list(convert_packet_header_to_dict(self.header).keys())
        self.assertEqual(expected_fields, actual_field)

    def test_year_is_converted_correctly(self):
        with self.subTest('test_year_has_fewer_than_two_digits'):
            self.header.time = UTCDateTime(year=51, month=1, day=1)
            expected_year = 51
            actual_year = convert_packet_header_to_dict(self.header)['year']
            self.assertEqual(actual_year, expected_year)
        with self.subTest('test_year_has_more_than_two_digits'):
            self.header.time = UTCDateTime(year=2352, month=1, day=1)
            expected_year = 52
            actual_year = convert_packet_header_to_dict(self.header)['year']
            self.assertEqual(actual_year, expected_year)
        with self.subTest('test_year_has_two_digits'):
            self.header.time = UTCDateTime(year=9, month=1, day=1)
            expected_year = 9
            actual_year = convert_packet_header_to_dict(self.header)['year']
            self.assertEqual(actual_year, expected_year)


class TestConvertWaveformPacketToObspyFormat(BaseTestCase):
    def setUp(self) -> None:
        TEST_DATA_DIR = Path(os.getcwd()).joinpath('tests/test_data')
        rt130_dir = TEST_DATA_DIR.joinpath(
            'RT130-sample/2017149.92EB/2017150/92EB'
        )
        file = rt130_dir.joinpath('1/000000015_0036EE80').as_posix()
        self.packet = read_rt130_file(file, unpacker)[0]

    def test_all_needed_fields_are_available(self):
        converted_packet = convert_waveform_packet_to_obspy_format(
            self.packet, unpacker
        )

        self.assertEqual(len(converted_packet), len(PACKET_FINAL_DTYPE))


class TestConvertSOHPacketToObspyFormat(BaseTestCase):
    def setUp(self) -> None:
        TEST_DATA_DIR = Path(os.getcwd()).joinpath('tests/test_data')
        rt130_dir = TEST_DATA_DIR.joinpath(
            'RT130-sample/2017149.92EB/2017150/92EB'
        )
        file = rt130_dir.joinpath('0/000000000_00000000').as_posix()
        self.packet = read_rt130_file(file, unpacker)[0]

    def test_all_needed_fields_are_available(self):
        converted_packet = convert_soh_packet_to_obspy_format(
            self.packet, unpacker
        )

        self.assertEqual(len(converted_packet),
                         len(soh_packet.PACKET_FINAL_DTYPE))
